@@ -1,6 +1,6 @@
 package Git::Raw::Commit;
 {
-  $Git::Raw::Commit::VERSION = '0.27';
+  $Git::Raw::Commit::VERSION = '0.28';
 }
 
 use strict;
@@ -14,7 +14,7 @@ Git::Raw::Commit - Git commit class
 
 =head1 VERSION
 
-version 0.27
+version 0.28
 
 =head1 SYNOPSIS
 
@@ -55,11 +55,13 @@ B<WARNING>: The API of this module is unstable and may change without warning
 
 =head1 METHODS
 
-=head2 create( $repo, $msg, $author, $committer, [@parents], $tree )
+=head2 create( $repo, $msg, $author, $committer, [@parents], $tree [, $update_ref ] )
 
 Create a new commit given a message, two L<Git::Raw::Signature> (one is the
 commit author and the other the committer), a list of parent commits and a
-L<Git::Raw::Tree>.
+L<Git::Raw::Tree>.  If C<$update_ref> is provided and is defined, the reference with
+the corresponding name is automatically updated or created.  If C<$update_ref> is
+C<undef>, no reference is updated.  If C<$update_ref> is not provided, "HEAD" is updated.
 
 =head2 lookup( $repo, $id )
 
